/*
 * k9tivo
 * 
 * k9tivo is a simple NTP broadcast client based on the k9linux client src.
 * 
 * (C) Copyright H.C.Mingham-Smith Ltd. 1999.  All rights reserved. Modified for
 * TiVo use by Warren Toomey 2004.
 * 
 * You have a royalty-free right to use, modify, reproduce and distribute the
 * Sample Files (and/or any modified version) in any way you find useful,
 * provided that you agree that H.C.Mingham-Smith Ltd. has no warranty
 * obligations or liability for any Executable Files generated from this
 * source code.
 * 
 * Implements a basic NTP broadcast listener
 * 
 * Command line options are
 * 
 * -d	Run as debug (don't run in background, set the time, show trace)
 * -o 	Run until time received then die
 * -b	Beep when message received
 * -m   Bind to the given multicast address
 * -s 	Log to syslog
 * 
 * Build this source as follows
 * 
 * cc k9tivo.c -o k9tivo
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<syslog.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/time.h>
#include	"tivotimelib.h"

#define	TRUE	1
#define	FALSE	0

#ifdef __alpha__
#define long int
#endif


/* NTP packet (no authentication) */

struct ntppkt {
  unsigned char flags[4];
  unsigned long rootdelay;
  unsigned long rootdispersion;
  unsigned long refid;
  unsigned long reftimehi;
  unsigned long reftimelo;
  unsigned long orghi;
  unsigned long orglo;
  unsigned long rechi;
  unsigned long reclo;
  unsigned long xmthi;
  unsigned long xmtlo;
};

/* Last server, stratum, time used */

struct last {
  int stratum;
  unsigned long addr;
  time_t time;
}   last;


int
main(int argc, char **argv)
{
  extern int verbose;
  struct sockaddr_in srv_addr;
  struct sockaddr_in remoteAddr;
  struct ip_mreq mreq;
  struct ntppkt ipkt;
  struct timeval tv;
  time_t secs;
  time_t subsecs;
  int stratum;
  long offset;

  int remoteAddrLength = sizeof(remoteAddr);
  int mastersd;
  int i;

  char debug = FALSE;
  char runonce = FALSE;
  char beep = FALSE;
  char dosyslog = FALSE;
  char *multicast_address = NULL;

  while ((i = getopt(argc, argv, "dbosm:")) != EOF) {
    switch (i) {
    case 'd':
      debug = TRUE;
      verbose = 1;
      break;

    case 'o':
      runonce = TRUE;
      break;

    case 'b':
      beep = TRUE;
      break;

    case 's':
      dosyslog = TRUE;
      break;

    case 'm':
      multicast_address = optarg;
      break;

    default:
      printf("Usage: k9tivo [-b][-d][-o][-s][-m multicast_address]\n");
      exit(1);
    }
  }


  if (debug) {
    printf("K9 V1.3  H.C. Mingham-Smith Ltd. 1998-2000\n");
  } else {
    daemon(0,0);
    if (dosyslog) {
      openlog("k9tivo", LOG_PID, LOG_DAEMON);
      syslog(LOG_INFO, "K9 V1.3  H.C. Mingham-Smith Ltd. 1998-2000\n");
    }
  }

  if ((mastersd = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("Couldn't create socket");
    exit(1);
  }

  srv_addr.sin_family = AF_INET;
  srv_addr.sin_addr.s_addr = INADDR_ANY;
  srv_addr.sin_port = htons(123);

  if (bind(mastersd, (struct sockaddr *) & srv_addr, sizeof(srv_addr)) == -1) {
    perror("Couldn't bind socket");
    exit(1);
  }

  if (multicast_address) {	/* Join Multicast group */
    mreq.imr_multiaddr.s_addr = inet_addr(multicast_address);
    mreq.imr_interface.s_addr = INADDR_ANY;

    if (setsockopt(mastersd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq)) == -1) {
      perror("Couldn't add membership of multicast group");
    }
  }

  while (1) {
    if (recvfrom(mastersd, (char *)&ipkt, sizeof(ipkt), 0, (struct sockaddr *) & remoteAddr, &remoteAddrLength) != -1) {

      if ((ipkt.flags[0] & 7) != 5)	/* Loop if it is not a broadcast */
	continue;

      secs = ntohl(ipkt.xmthi);
      secs = secs - 2208988800U;	/* Seconds since 1900 */
      subsecs = ntohl(ipkt.xmtlo);
      stratum = ipkt.flags[1];

      if (debug) {
	printf("Time broadcast received from %s\n",
	       inet_ntoa(remoteAddr.sin_addr));
	printf("Ver: %d Stratum: %d Mode: %d Poll:%d Precision %d\n", ((ipkt.flags[0] & 0x38) >> 3),
	       ipkt.flags[1], (ipkt.flags[0] & 0x07), ipkt.flags[2],
	       (char)ipkt.flags[3]);
	printf("UTC Time is %s",
	       asctime(gmtime(&secs)));
      }

      /* Only set the system time if the stratum is valid */
      if ((stratum <= 0) && (stratum > 16))
	continue;

      /*
       * Only set the system time if it is from our selected server or the
       * stratum is higher than our server or it has been too long since a
       * messge was received.
       */
      if ((last.addr == remoteAddr.sin_addr.s_addr) ||
	  (last.stratum > stratum) ||
	  (abs(secs - last.time) > 600)) {

	offset = get_timezone_offset(0);
	if (offset==-1) {
	  if (debug) printf("PANIC: couldn't find TiVo's time zone\n");
	  continue;
	}
	if (debug) printf("UTC to local time offset is %ld\n",offset);

	secs+= offset;
	if (debug)
	  printf("Setting the local time to %s", asctime(gmtime(&secs)));
	else
	   syslog(LOG_INFO, "Setting the local time to %s",
						asctime(gmtime(&secs)));

	tv.tv_sec = secs;
	/*
	 * The next line was originally this: tv.tv_usec=(unsigned long)
	 * 1000000 * ((double)subsecs/(double)0x100000000); but the TiVo
	 * doesn't do floating-point. We should be dividing by 4294.967296,
	 * but 4295 is close enough.
	 */
	tv.tv_usec = (unsigned long)subsecs / 4295U;
	if (settimeofday(&tv, NULL) == -1) {
	  perror("Couldn't set the time");
	}
	last.addr = remoteAddr.sin_addr.s_addr;
	last.time = secs;
	last.stratum = stratum;

	if (beep) {		/* Beep when we get a message */
	  write(1, "\a", 1);
	}
	if (runonce) {		/* exit when we get a message */
	  break;
	}
      }
    }
  }
  return (0);
}
