/*
 * ntpdate.c: Program to obtain the UTC time from a time server,
 * and to print out a localtime strings which is suitable for use by settime.
 *
 * NOTE: We actually don't use NTP. The name is simply to keep the TiVo
 * happy.
 *
 * Copyright (c) 2004 Warren Toomey and Andrew Tridgell.
 * The program is placed under the GPL.
 */

#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tivotimelib.h"

extern int iicw1(int module, int subaddress, unsigned char v);

const char *DaylightSavingsMarker = "/var/tmp/dst_lastmodified";
static const char *Version= "$Revision: 1.17 $";
static int set_the_time=1;		/* We want to set the time */
extern int verbose;


/* Open a socket to a tcp remote host with the specified port */
static int open_socket_out(const char *host, int port)
{
  struct sockaddr_in sock_out;
  int res;
  struct in_addr addr;

  if (verbose)
    fprintf(stderr, "Connecting to time server %s\n",host);
  res = socket(PF_INET, SOCK_STREAM, 0);
  if (res == -1) {
    if (verbose)
      fprintf(stderr, "Failed to obtain a socket: %s\n", strerror(errno));
    else
      syslog(LOG_INFO, "Failed to obtain a socket: %s\n", strerror(errno));
    return(-1);
  }

  inet_aton(host, &addr);
  sock_out.sin_addr = addr;
  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *) & sock_out, sizeof(sock_out)) != 0) {
    close(res);
    if (verbose)
      fprintf(stderr, "Failed to connect to %s: %s\n", host, strerror(errno));
    else
      syslog(LOG_INFO, "Failed to connect to %s: %s\n", host, strerror(errno));
    return(-1);
  }
  return(res);
}


/*
 * Convert incoming number into two hex nybbles. This assumes that
 * the number is in the range 0 to 99, but I don't check this at all.
 */
unsigned char hexify(int val)
{
  unsigned char result;

  result = ((16 * (val / 10)) + (val % 10));
  return (result);
}

/*
 * Given the # seconds since 1970, set the TiVo's hardware clock.
 * We convert into yy/mm/dd hh/mm/ss fields, hexify each field,
 * and then write them to the M41T56 clock chip which lives on the
 * I2C bus at CLOCK_I2C_ADDR.
 */
void set_hardware_clock(time_t tim)
{
#define CLOCK_I2C_ADDR 0xd0
  struct tm *T;
  unsigned char hw_bytes[7];
  int i;

  T = gmtime(&tim);
  if (T == NULL) {
    fprintf(stderr, "Cannot convert time_t to struct tm\n");
    return;
  }

#ifdef DEBUG
  printf("gmtime gives    %d/%d/%d %02d:%02d:%02d  [%d]\n",
	 T->tm_year, T->tm_mon, T->tm_mday,
	 T->tm_hour, T->tm_min, T->tm_sec, T->tm_wday);
#endif

  hw_bytes[0] = hexify(T->tm_sec);
  hw_bytes[1] = hexify(T->tm_min);
  hw_bytes[2] = hexify(T->tm_hour);
  hw_bytes[3] = hexify(T->tm_wday);
  hw_bytes[4] = hexify(T->tm_mday);
  hw_bytes[5] = hexify(T->tm_mon + 1);		/* Convert 0-11 to 1-12 */
  hw_bytes[6] = hexify(T->tm_year - 100);	/* Make it years since 2000 */

#ifdef DEBUG
  printf("sending to clock %02x/%02x/%02x %02x:%02x:%02x  [%x]\n",
	 hw_bytes[6], hw_bytes[5], hw_bytes[4],
	 hw_bytes[2], hw_bytes[1], hw_bytes[0], hw_bytes[3]);
#endif

  /* Now write to the I2C device */
  for (i = 0; i < 7; i++)
    iicw1(CLOCK_I2C_ADDR, i, hw_bytes[i]);
}

/* Simply open /var/timestamp for writing
 * is enough to "touch" it.
 */
static void 
touch_vartimestamp(void)
{
  int fd;
  fd=open("/var/timestamp", O_WRONLY|O_CREAT);
  close(fd);
}

static void 
tivo_time(int s, long ofs)
{
  struct timeval tv;
  unsigned long t;
  if (read(s, &t, sizeof(t)) != 4) {
    if (verbose)
      fprintf(stderr, "Unable to read time from server: %s\n",strerror(errno));
    else
      syslog(LOG_INFO, "Unable to read time from server: %s\n",strerror(errno));
    exit(1);
  }

  t = ntohl(t);
  t += (ofs - 2208988800U);
  if (set_the_time) {
    if (verbose)
      printf("Setting the local time to %s", asctime(gmtime(&t)));
    syslog(LOG_INFO, "Setting the local time to %s",
                                                asctime(gmtime(&t)));
    tv.tv_sec = t;
    tv.tv_usec = 0;
    if (settimeofday(&tv, NULL) == -1) {
      if (verbose)
        printf("Couldn't settimeofday\n");
      else
        syslog(LOG_INFO, "Couldn't settimeofday\n");
    }
    set_hardware_clock(t);
  }
}

void do_daylight_saving_change(long prevoffset, long offset)
{
  FILE *zout;
  struct stat sb;
  time_t secs;
  struct timeval tv;
  int err;

  if (verbose)
    fprintf(stderr, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);
  else
    syslog(LOG_INFO, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);

  /* We have to be careful and not adjust the clock several times a la
   * Windoze. If the DaylightSavingsMarker is more than a day old, then
   * make the change. Otherwise, don't make the change.
   */
  secs = time(NULL);
  err=stat(DaylightSavingsMarker, &sb);
  if ((err==0) && ((secs - sb.st_mtime)<86400)) {
    if (verbose)
      fprintf(stderr, "Not doing daylight saving change: I already did it\n");
    else
      syslog(LOG_INFO, "Not doing daylight saving change: I already did it\n");
    exit(0);
  }

  /* Make the change, touch the file. */
  zout= fopen(DaylightSavingsMarker, "w");
  if (zout==NULL) exit(1);
  fprintf(zout, "Daylight saving change: offset %ld -> %ld in secs\n",
                                                        prevoffset,offset);
  fclose(zout);
  
  secs= secs + (offset - prevoffset);
  
  if (set_the_time) {
    if (verbose)
      printf("Setting the Local time to %s", asctime(gmtime(&secs)));
    syslog(LOG_INFO, "Setting the local time to %s",
                                                asctime(gmtime(&secs)));
    tv.tv_sec = secs;
    tv.tv_usec = 0;
    if (settimeofday(&tv, NULL) == -1) {
      if (verbose)
        printf("Couldn't settimeofday\n");
      else
        syslog(LOG_INFO, "Couldn't settimeofday\n");
    }
  }
  exit(0);
}

int 
main(int argc, char *argv[])
{
  char *host;
  int ch, sock;
  int servernum;
  long offset,prevoffset;

  if (argc < 2) {
    fprintf(stderr, "ntpdate [-B] [-b] [d] [-q] [-v] server [server ...]\n");
    exit(1);
  }

  openlog("ntpdate", LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "fake ntpdate has started");

  while ((ch = getopt(argc, argv, "Bbdqv")) != -1)
    switch (ch) {
      case 'B':
      case 'b': break;
      case 'd':
      case 'q': set_the_time=0; break;
      case 'v': verbose=1; break;
    }
  argc -= optind;
  argv += optind;

  offset = get_timezone_offset(0);
  if (offset==-1) {
    if (verbose) 
      printf("PANIC: couldn't find TiVo's time zone\n");
    syslog(LOG_INFO, "PANIC: couldn't find TiVo's time zone\n");
    exit(1);
  }

  for (servernum=0; servernum<argc; servernum++) {
    host = argv[servernum];
    sock = open_socket_out(host, 37);
    if (sock<0) continue;
    syslog(LOG_INFO, "Connected to time server %s", host);
    tivo_time(sock, offset);
    close(sock);
    closelog();
    touch_vartimestamp();
    exit(0);
  }

  /* We failed to connect to any servers */
  if (verbose)
    fprintf(stderr, "Failed to connect to any of the time servers\n");
  else
    syslog(LOG_INFO, "Failed to connect to any of the time servers\n");

  /* Do nothing if the offset now and 2 hours ago are the same */
  prevoffset = get_timezone_offset(2);
  if (offset==prevoffset) exit(1);

  do_daylight_saving_change(prevoffset,offset);
  touch_vartimestamp();
  exit(0);
}
