/*
 * tivotimelib.c: Routines to determine the correct timezone for this TiVo
 * given its timezone, and to calculate the offset for the timezone from
 * UTC.
 *
 * Copyright (c) 2004 Warren Toomey.
 * The code is placed under the GPL.
 */

#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include "tivotimelib.h"


/* External files and programs */
const char *PostcodeZones = "/etc/postcodezones";
const char *Get_Postalcode = "/var/hack/bin/get_postalcode";
static const char *Version= "$Revision: 1.3 $";

#define LIST_SIZE 100
char *Postcode[LIST_SIZE];
char *Timezone[LIST_SIZE];
int Listsize=0;
int verbose=0;


static void read_postcode_zones(void)
{
  FILE *zonefile;
#define BUF_SIZE 200
  char buf[BUF_SIZE];
  char *tab;

  /* Open the file */
  zonefile = fopen(PostcodeZones, "r");
  if (zonefile == NULL) {
    if (verbose)
      fprintf(stderr, "Can't open %s: %s\n", PostcodeZones,  strerror(errno));
    else
      syslog(LOG_INFO, "Can't open %s: %s\n", PostcodeZones,  strerror(errno));
    exit(1);
  }
  /* Get each postcode and timezone */
  for (Listsize = 0; Listsize < LIST_SIZE;) {
    if (fgets(buf, BUF_SIZE - 1, zonefile) == NULL) break;

    /* Ignore comment lines */
    if (buf[0] == '#') continue;

    /* Trash the newline character */
    buf[strlen(buf) - 1] = '\0';

    /* Find the single tab in the line. Skip line if none. */
    tab = strchr(buf, '\t');
    if (tab == NULL) continue;

    /* Split the string at this point */
    *tab = '\0'; tab++;

    /* Save copies of the strings */
    Postcode[Listsize] = strdup(buf);
    Timezone[Listsize] = strdup(tab);
    Listsize++;
  }
  fclose(zonefile);
}

static char *get_postcode(void)
{
  FILE *postcodepipe;
  char postcode[6] = "     ";
  char junk[100];
  int count, err;

  /* Obtain our postcode. We loop a few times in case */
  /* the MFS is in use already. */
  for (count=0; count<4; count++) {
    postcodepipe = popen(Get_Postalcode, "r");
    if (postcodepipe == NULL) {
      if (verbose)
        fprintf(stderr, "Failed to open get_postalcode pipe: %s\n",
							strerror(errno));
      else
        syslog(LOG_INFO, "Failed to open get_postalcode pipe: %s\n",
							strerror(errno));
      exit(1);
    }

    err = fread(postcode, 1, 5, postcodepipe);
    if (err != 5) {
      if (verbose)
        fprintf(stderr, "Failed to read from get_postalcode pipe: %s\n",
							strerror(errno));
        else
        syslog(LOG_INFO, "Failed to read from get_postalcode pipe: %s\n",
							strerror(errno));
      exit(1);
    }

    /* Read the rest of the pipe and close the input pipe */
    while (!feof(postcodepipe)) fread(junk, 1, 90, postcodepipe);
    fclose(postcodepipe);

    /* See if we have to retry the operation */
    if (!strcmp(postcode, "retry")) {
      sleep(2); continue;
    }

    return (strdup(postcode));
  }
  if (verbose)
    fprintf(stderr, "Failed %d times to get_postalcode\n", count);
  else
    syslog(LOG_INFO, "Failed %d times to get_postalcode\n", count);
  exit(1);
}

void set_tz_variable(void)
{
  int i, err;
  char *pcode;

  if (Listsize==0) read_postcode_zones();
  pcode = get_postcode();
  if (pcode == NULL) {
    if (verbose)
      fprintf(stderr, "Null postcode in set_tz_variable()\n");
    else
      syslog(LOG_INFO, "Null postcode in set_tz_variable()\n");
    exit(1);
  }
  if (verbose)
    fprintf(stderr, "Our postcode is %s. ", pcode);

  /* Walk the list to find a matching postcode */
  for (i = 0; i < Listsize; i++) {
    if (!strcmp(Postcode[i], pcode)) {
      if (verbose)
	fprintf(stderr, "Found matching timezone %s\n", Timezone[i]);
      err = setenv("TZ", Timezone[i], 1);
      if (err == -1) {
	if (verbose)
	  fprintf(stderr, "Error setting the TZ variable: %s\n",
							strerror(errno));
 	else
	  syslog(LOG_INFO, "Error setting the TZ variable: %s\n",
							strerror(errno));
	exit(1);
      }
      return;
    }
  }
  if (verbose)
    fprintf(stderr, "No matching timezone for postcode %s\n", pcode);
  else
    syslog(LOG_INFO, "No matching timezone for postcode %s\n", pcode);
  exit(1);
}

/* Returns -1 on errors */
long get_timezone_offset(int hoursago)
{
  time_t secs;
  struct tm *tim;
  long offset;

  /* Set the TV environment variable if it is not already set */
  if (getenv("TZ")==NULL)
     set_tz_variable();
  if (getenv("TZ")==NULL)
     return(-1);
  /* Get the current local time */
  secs = time(NULL);
  /* Subtract hoursago */
  secs = secs - (hoursago*3600);
  /* Localtime expects the number of GMT seconds */
  tim = localtime(&secs);
  offset = tim->tm_gmtoff;
  /* So now that we have the offset, we subtract it from secs */
  /* and get a more accurate offset */
  secs -= (time_t) offset;
  tim = localtime(&secs);
  offset = tim->tm_gmtoff;
  if (verbose && (hoursago==0))
    fprintf(stderr, "Offset %d for timezone %s, daylight saving is %d\n",
				 (int)offset, tim->tm_zone, tim->tm_isdst);
  return (offset);
}
